## vim: filetype=makoada

<% list_kind_range = ctx.generic_list_type.ada_kind_range_name %>

with Ada.Containers.Hashed_Maps;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Strings.Unbounded.Hash;

with ${ada_lib_name}.Implementation;
use ${ada_lib_name}.Implementation;

<%def name="return_program_error()">
   pragma Warnings (Off, "value not in range of type");
   return (raise Program_Error);
   pragma Warnings (On, "value not in range of type");
</%def>

package body ${ada_lib_name}.Introspection is

   type String_Access is access constant String;
   type String_Array is array (Positive range <>) of String_Access;

   ------------------------------
   -- Syntax field descriptors --
   ------------------------------

   type Syntax_Field_Descriptor (Name_Length : Natural) is record
      Field_Type : Node_Type_Id;
      Name       : String (1 .. Name_Length);
   end record;
   --  General description of a field (independent of field implementations)

   type Syntax_Field_Descriptor_Access is
      access constant Syntax_Field_Descriptor;

   --  Descriptors for syntax fields

   % for f in ctx.sorted_parse_fields:
      <% name = f._name.lower %>
      Desc_For_${f.introspection_enum_literal} : aliased constant
         Syntax_Field_Descriptor := (
            Name_Length => ${len(name)},
            Field_Type  => ${f.type.introspection_name},
            Name        => ${string_repr(name)}
         );
   % endfor

   Syntax_Field_Descriptors : constant
      array (Field_Reference) of Syntax_Field_Descriptor_Access := (
      % if ctx.sorted_parse_fields:
         ${', '.join("{name} => Desc_For_{name}'Access"
                     .format(name=f.introspection_enum_literal)
                     for f in ctx.sorted_parse_fields)}
      % else:
         Field_Reference => <>
      % endif
   );

   --------------------------
   -- Property descriptors --
   --------------------------

   type Property_Descriptor (
      Name_Length : Natural;
      --  Length of the proprety name

      Arity : Natural
      --  Number of arguments this property takes (exclude the Self argument)
   )
   is record
      Name : String (1 .. Name_Length);
      --  Lower-case name for this property

      Return_Type : Value_Constraint;
      --  Return type for this property

      Argument_Types : Value_Constraint_Array (1 .. Arity);
      --  Types of the arguments that this property takes

      Argument_Names : String_Array (1 .. Arity);
      --  Lower-case names for arguments that this property takes
   end record;

   type Property_Descriptor_Access is access constant Property_Descriptor;

   --  Descriptors for properties

   <%
      # First, generate constant for argument names, so that we can refer to
      # them from property descriptors.
      names = set()
      for p in ctx.sorted_properties:
         for arg in p.arguments:
            names.add(arg.name.lower)
   %>
   % for n in sorted(names):
   Name_For_${n} : aliased constant String := ${string_repr(n)};
   % endfor

   % for p in ctx.sorted_properties:
      <% name = p._name.lower %>
      Desc_For_${p.introspection_enum_literal} : aliased constant
         Property_Descriptor := (
            Name_Length => ${len(name)},
            Arity       => ${len(p.arguments)},

            Name => ${string_repr(name)},

            Return_Type    => ${p.type.introspection_constraint},
            Argument_Types => (
               % if p.arguments:
                  ${', '.join('{} => {}'
                              .format(i, arg.type.introspection_constraint)
                              for i, arg in enumerate(p.arguments, 1))}
               % else:
                  1 .. 0 => <>
               % endif
            ),
            Argument_Names => (
               % if p.arguments:
                  ${', '.join("{} => Name_For_{}'Access"
                              .format(i, arg.name.lower)
                              for i, arg in enumerate(p.arguments, 1))}
               % else:
                  1 .. 0 => <>
               % endif
            )
         );
   % endfor

   % if ctx.sorted_properties:
      Property_Descriptors : constant
         array (Property_Reference) of Property_Descriptor_Access := (
         % if ctx.sorted_properties:
            ${', '.join("Desc_For_{}'Access"
                        .format(p.introspection_enum_literal)
                        for p in ctx.sorted_properties)}
         % else:
            Property_Reference => <>
         % endif
      );
   % endif

   ---------------------------
   -- Node type descriptors --
   ---------------------------

   type Node_Field_Descriptor (Is_Abstract_Or_Null : Boolean) is record
      Field : Field_Reference;
      --  Reference to the field this describes

      --  Only non-null concrete fields are assigned an index

      case Is_Abstract_Or_Null is
         when False =>
            Index : Positive;
            --  Index for this field

         when True =>
            null;
      end case;
   end record;
   --  Description of a field as implemented by a specific node

   type Node_Field_Descriptor_Access is access constant Node_Field_Descriptor;
   type Node_Field_Descriptor_Array is
      array (Positive range <>) of Node_Field_Descriptor_Access;

   type Node_Type_Descriptor
     (Is_Abstract       : Boolean;
      Derivations_Count : Natural;
      Fields_Count      : Natural;
      Properties_Count  : Natural)
   is record
      Base_Type : Any_Node_Type_Id;
      --  Reference to the node type from which this derives

      Derivations : Node_Type_Id_Array (1 .. Derivations_Count);
      --  List of references for all node types that derives from this

      DSL_Name : Unbounded_String;
      --  Name for this type in the Langkit DSL

      Inherited_Fields : Natural;
      --  Number of syntax field inherited from the base type

      Fields : Node_Field_Descriptor_Array (1 .. Fields_Count);
      --  For regular node types, list of syntax fields that are specific to
      --  this derivation (i.e. excluding fields from the base type).

      Properties : Property_Reference_Array (1 .. Properties_Count);
      --  List of properties that this node provides that are specific to this
      --  derivation (i.e. excluding fields from the base type).

      --  Only concrete nodes are assigned a node kind

      case Is_Abstract is
         when False =>
            Kind : ${root_node_kind_name};
            --  Kind corresponding this this node type

         when True =>
            null;
      end case;
   end record;

   type Node_Type_Descriptor_Access is access constant Node_Type_Descriptor;

   --  Descriptors for node types and their syntax fields

   % for n in ctx.astnode_types:
   <%
      fields = n.get_parse_fields(include_inherited=False)
      properties = n.get_properties(
         predicate=lambda p: p.is_public and not p.overriding,
         include_inherited=False)

      if n.is_root_node:
         inherited_fields = []
      else:
         inherited_fields = n.base.get_parse_fields(
            predicate=lambda f: not (f.abstract or f.null),
            include_inherited=True)
   %>

   % for f in fields:
   ${f.name}_For_${n.kwless_raw_name} : aliased constant Node_Field_Descriptor
   := (
      Is_Abstract_Or_Null => ${f.abstract or f.null},
      Field               => ${(f.base or f).introspection_enum_literal}

      % if not f.abstract and not f.null:
         , Index => ${f.index + 1}
      % endif
   );
   % endfor

   Desc_For_${n.kwless_raw_name} : aliased constant Node_Type_Descriptor := (
      Is_Abstract       => ${n.abstract},
      Derivations_Count => ${len(n.subclasses)},
      Fields_Count      => ${len(fields)},
      Properties_Count  => ${len(properties)},

      Base_Type   => ${n.base.introspection_name if n.base else 'None'},
      Derivations =>
         ${('({})'.format(', '.join(
            '{} => {}'.format(i, child.introspection_name)
            for i, child in enumerate(n.subclasses, 1)
         )) if n.subclasses else '(1 .. 0 => <>)')},

      DSL_Name => To_Unbounded_String ("${n.dsl_name}"),

      Inherited_Fields => ${len(inherited_fields)},
      Fields           => (
         % if fields:
            ${', '.join("{} => {}_For_{}'Access"
                        .format(i, f.name, n.kwless_raw_name)
                        for i, f in enumerate(fields, 1))}
         % else:
            1 .. 0 => <>
         % endif
      ),

      Properties => (
         % if properties:
            ${', '.join("{} => {}".format(i, p.introspection_enum_literal)
                        for i, p in enumerate(properties, 1))}
         % else:
            1 .. 0 => <>
         % endif
      )

      % if not n.abstract:
      , Kind => ${n.ada_kind_name}
      % endif
   );
   % endfor

   Node_Type_Descriptors : constant
      array (Node_Type_Id) of Node_Type_Descriptor_Access
   := (${', '.join("Desc_For_{}'Access".format(n.kwless_raw_name)
                   for n in ctx.astnode_types)});

   ----------------------
   -- Various mappings --
   ----------------------

   package Node_Type_Id_Maps is new Ada.Containers.Hashed_Maps
     (Key_Type        => Unbounded_String,
      Element_Type    => Node_Type_Id,
      Equivalent_Keys => "=",
      Hash            => Hash);

   DSL_Name_To_Node_Type : Node_Type_Id_Maps.Map;
   --  Lookup table for DSL names to node type references. Created at
   --  elaboration time and never updated after.

   Kind_To_Id : constant array (${root_node_kind_name}) of Node_Type_Id := (
      ${', '.join('{n.ada_kind_name} => {n.introspection_name}'.format(n=n)
                  for n in ctx.astnode_types
                  if not n.abstract)}
   );

   -------------
   -- Helpers --
   -------------

   function Fields
     (Id : Node_Type_Id; Concrete_Only : Boolean) return Field_Reference_Array;
   --  Return the list of fields associated to Id. If Concrete_Only is true,
   --  collect only non-null and concrete fields. Otherwise, collect all
   --  fields.

   function Allocate (Kind : Value_Kind) return Value_Type;
   --  Allocate a polymorphic value of the given kind

   --------------
   -- DSL_Name --
   --------------

   function DSL_Name (Id : Node_Type_Id) return String is
   begin
      return To_String (Node_Type_Descriptors (Id).DSL_Name);
   end DSL_Name;

   ---------------------
   -- Lookup_DSL_Name --
   ---------------------

   function Lookup_DSL_Name (Name : String) return Any_Node_Type_Id is
      use Node_Type_Id_Maps;

      Position : constant Cursor :=
         DSL_Name_To_Node_Type.Find (To_Unbounded_String (Name));
   begin
      if Has_Element (Position) then
         return Element (Position);
      else
         return None;
      end if;
   end Lookup_DSL_Name;

   -----------------
   -- Is_Abstract --
   -----------------

   function Is_Abstract (Id : Node_Type_Id) return Boolean is
   begin
      return Node_Type_Descriptors (Id).Is_Abstract;
   end Is_Abstract;

   --------------
   -- Kind_For --
   --------------

   function Kind_For (Id : Node_Type_Id) return ${root_node_kind_name} is
      Desc : Node_Type_Descriptor renames Node_Type_Descriptors (Id).all;
   begin
      if Desc.Is_Abstract then
         raise Constraint_Error with "trying to get kind for abstract node";
      end if;
      return Desc.Kind;
   end Kind_For;

   -----------------
   -- Id_For_Kind --
   -----------------

   function Id_For_Kind (Kind : ${root_node_kind_name}) return Node_Type_Id is
   begin
      return Kind_To_Id (Kind);
   end Id_For_Kind;

   ------------------
   -- Is_Root_Node --
   ------------------

   function Is_Root_Node (Id : Node_Type_Id) return Boolean is
   begin
      return Id = ${T.root_node.introspection_name};
   end Is_Root_Node;

   ---------------
   -- Base_Type --
   ---------------

   function Base_Type (Id : Node_Type_Id) return Node_Type_Id is
   begin
      if Is_Root_Node (Id) then
         raise Constraint_Error with "trying to get base type of root node";
      end if;
      return Node_Type_Descriptors (Id).Base_Type;
   end Base_Type;

   -------------------
   -- Derived_Types --
   -------------------

   function Derived_Types (Id : Node_Type_Id) return Node_Type_Id_Array is
   begin
      return Node_Type_Descriptors (Id).Derivations;
   end Derived_Types;

   ---------------------
   -- Is_Derived_From --
   ---------------------

   function Is_Derived_From (Id, Parent : Node_Type_Id) return Boolean is
      Cursor : Any_Node_Type_Id := Id;
   begin
      while Cursor /= None loop
         if Cursor = Parent then
            return True;
         end if;

         Cursor := Node_Type_Descriptors (Cursor).Base_Type;
      end loop;
      return False;
   end Is_Derived_From;

   ----------
   -- Kind --
   ----------

   function Kind (Self : Value_Type) return Value_Kind is
   begin
      return Self.Value.Value.Kind;
   end Kind;

   --------------
   -- Allocate --
   --------------

   function Allocate (Kind : Value_Kind) return Value_Type is
      Result : Any_Value_Type;
   begin
      Result.Value.Value := new Value_Record (Kind);
      Result.Value.Value.Ref_Count := 1;
      return Result;
   end Allocate;

   ----------------
   -- As_Boolean --
   ----------------

   function As_Boolean (Self : Value_Type) return Boolean is
   begin
      return Self.Value.Value.Boolean_Value;
   end As_Boolean;

   --------------------
   -- Create_Boolean --
   --------------------

   function Create_Boolean (Value : Boolean) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Boolean_Value) do
         Result.Value.Value.Boolean_Value := Value;
      end return;
   end Create_Boolean;

   ----------------
   -- As_Integer --
   ----------------

   function As_Integer (Self : Value_Type) return Integer is
   begin
      return Self.Value.Value.Integer_Value;
   end As_Integer;

   --------------------
   -- Create_Integer --
   --------------------

   function Create_Integer (Value : Integer) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Integer_Value) do
         Result.Value.Value.Integer_Value := Value;
      end return;
   end Create_Integer;

   --------------------
   -- As_Big_Integer --
   --------------------

   function As_Big_Integer (Self : Value_Type) return Big_Integer is
   begin
      return Result : Big_Integer do
         Result.Set (Self.Value.Value.Big_Integer_Value);
      end return;
   end As_Big_Integer;

   ------------------------
   -- Create_Big_Integer --
   ------------------------

   function Create_Big_Integer (Value : Big_Integer) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Big_Integer_Value) do
         Result.Value.Value.Big_Integer_Value.Set (Value);
      end return;
   end Create_Big_Integer;

   ------------------
   -- As_Character --
   ------------------

   function As_Character (Self : Value_Type) return Character_Type is
   begin
      return Self.Value.Value.Character_Value;
   end As_Character;

   ----------------------
   -- Create_Character --
   ----------------------

   function Create_Character (Value : Character_Type) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Character_Value) do
         Result.Value.Value.Character_Value := Value;
      end return;
   end Create_Character;

   --------------
   -- As_Token --
   --------------

   function As_Token (Self : Value_Type) return Token_Reference is
   begin
      return Self.Value.Value.Token_Value;
   end As_Token;

   ------------------
   -- Create_Token --
   ------------------

   function Create_Token (Value : Token_Reference) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Token_Value)
      do
         Result.Value.Value.Token_Value := Value;
      end return;
   end Create_Token;

   -----------------------
   -- As_Unbounded_Text --
   -----------------------

   function As_Unbounded_Text (Self : Value_Type) return Unbounded_Text_Type is
   begin
      return Self.Value.Value.Unbounded_Text_Value;
   end As_Unbounded_Text;

   ---------------------------
   -- Create_Unbounded_Text --
   ---------------------------

   function Create_Unbounded_Text
     (Value : Unbounded_Text_Type) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Unbounded_Text_Value)
      do
         Result.Value.Value.Unbounded_Text_Value := Value;
      end return;
   end Create_Unbounded_Text;

   ----------------------
   -- As_Analysis_Unit --
   ----------------------

   function As_Analysis_Unit (Self : Value_Type) return Analysis_Unit is
   begin
      return Self.Value.Value.Analysis_Unit_Value;
   end As_Analysis_Unit;

   --------------------------
   -- Create_Analysis_Unit --
   --------------------------

   function Create_Analysis_Unit (Value : Analysis_Unit) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Analysis_Unit_Value) do
         Result.Value.Value.Analysis_Unit_Value := Value;
      end return;
   end Create_Analysis_Unit;

   -------------
   -- As_Node --
   -------------

   function As_Node (Self : Value_Type) return ${root_entity.api_name} is
   begin
      return Self.Value.Value.Node_Value;
   end As_Node;

   -----------------
   -- Create_Node --
   -----------------

   function Create_Node
     (Value : ${root_entity.api_name}'Class) return Value_Type is
   begin
      return Result : constant Value_Type := Allocate (Node_Value) do
         Result.Value.Value.Node_Value := Value.As_${root_entity.api_name};
      end return;
   end Create_Node;

   % for enum_type in ctx.enum_types:
      function As_${enum_type.introspection_prefix}
        (Self : Value_Type) return ${enum_type.api_name} is
      begin
         return Self.Value.Value.${enum_type.introspection_kind};
      end As_${enum_type.introspection_prefix};

      function Create_${enum_type.introspection_prefix}
        (Value : ${enum_type.api_name}) return Value_Type is
      begin
         return Result : constant Value_Type := Allocate
           (${enum_type.introspection_kind})
         do
            Result.Value.Value.${enum_type.introspection_kind} := Value;
         end return;
      end Create_${enum_type.introspection_prefix};
   % endfor

   % for t in ctx.composite_types:
      % if t.exposed and not t.is_entity_type:
         function As_${t.introspection_prefix}
           (Self : Value_Type) return ${t.api_name} is
         begin
            % if t.is_array_type:
               ## If `t` is an array, first allocate the array and then
               ## initialize it one item at a time.
               return Result : ${t.api_name}
                 (Self.Value.Value.${t.introspection_kind}'Range)
               do
                  for I in Result'Range loop
                     ## Special case for big integer types: they are limited,
                     ## so we cannot use mere assignment.
                     % if t.element_type.is_big_integer_type:
                        Result (I).Set
                          (Self.Value.Value.${t.introspection_kind}.all (I));
                     % else:
                        Result (I) :=
                           Self.Value.Value.${t.introspection_kind}.all (I);
                     % endif
                  end loop;
               end return;

            % else:
               ## For other types, a mere assignment is fine
               return Self.Value.Value.${t.introspection_kind};
            % endif
         end As_${t.introspection_prefix};

         function Create_${t.introspection_prefix}
           (Value : ${t.api_name}) return Value_Type is
         begin
            return Result : constant Value_Type := Allocate
              (${t.introspection_kind})
            do
               % if t.is_array_type:
                  ## If `t` is an array, first allocate the array and then
                  ## initialize it one item at a time.
                  Result.Value.Value.${t.introspection_kind} :=
                     new ${t.api_name} (Value'Range);
                  for I in Value'Range loop
                     ## Special case for big integer types: they are limited,
                     ## so we cannot use mere assignment.
                     % if t.element_type.is_big_integer_type:
                        Result.Value.Value.${t.introspection_kind}
                           .all (I).Set (Value (I));
                     % else:
                        Result.Value.Value.${t.introspection_kind}.all (I) :=
                           Value (I);
                     % endif
                  end loop;

               % else:
                  ## For other types, a mere assignment is fine
                  Result.Value.Value.${t.introspection_kind} := Value;
               % endif
            end return;
         end Create_${t.introspection_prefix};
      % endif
   % endfor

   --------------
   -- DSL_Name --
   --------------

   function DSL_Name (Constraint : Value_Constraint) return String is
   begin
      <% basic_types = [T.Bool, T.Int, T.BigInt, T.Character, T.Token,
                        T.Symbol, T.AnalysisUnit] %>
      case Constraint.Kind is
         % for t in basic_types + ctx.enum_types + ctx.composite_types:
            % if t.exposed and not t.is_entity_type:
               when ${t.introspection_kind} =>
                  return "${t.dsl_name}";
            % endif
         % endfor

         when Node_Value =>
            return DSL_Name (Constraint.Node_Type);
      end case;
   end DSL_Name;

   ---------------
   -- Satisfies --
   ---------------

   function Satisfies
     (Value : Value_Type; Constraint : Value_Constraint) return Boolean is
   begin
      if Value.Value.Value.Kind /= Constraint.Kind then
         return False;
      end if;

      case Constraint.Kind is
         when Node_Value =>
            return Is_Derived_From
              (Id_For_Kind (Value.Value.Value.Node_Value.Kind),
               Constraint.Node_Type);

         when others =>
            return True;
      end case;
   end Satisfies;

   --------------------
   -- Node_Data_Name --
   --------------------

   function Node_Data_Name (Node_Data : Node_Data_Reference) return String is
   begin
      case Node_Data is
         when Field_Reference =>
            pragma Warnings (Off, "value not in range of type");
            return Field_Name (Node_Data);
            pragma Warnings (On, "value not in range of type");

         when Property_Reference =>
            return Property_Name (Node_Data);
      end case;
   end Node_Data_Name;

   --------------------
   -- Node_Data_Type --
   --------------------

   function Node_Data_Type
     (Node_Data : Node_Data_Reference) return Value_Constraint is
   begin
      case Node_Data is
         when Field_Reference =>
            pragma Warnings (Off, "value not in range of type");
            return (Kind => Node_Value, Node_Type => Field_Type (Node_Data));
            pragma Warnings (On, "value not in range of type");

         when Property_Reference =>
            return Property_Return_Type (Node_Data);
      end case;
   end Node_Data_Type;

   --------------------
   -- Eval_Node_Data --
   --------------------

   function Eval_Node_Data
     (Node      : ${T.entity.api_name}'Class;
      Node_Data : Node_Data_Reference;
      Arguments : Value_Array) return Value_Type is
   begin
      case Node_Data is
         when Field_Reference =>
            if Arguments'Length > 0 then
               raise Node_Data_Evaluation_Error with "fields take no argument";
            end if;
            pragma Warnings (Off, "value not in range of type");
            return Create_Node (Eval_Field (Node, Node_Data));
            pragma Warnings (On, "value not in range of type");

         when Property_Reference =>
            return Eval_Property (Node, Node_Data, Arguments);
      end case;
   end Eval_Node_Data;

   ----------------------
   -- Lookup_Node_Data --
   ----------------------

   function Lookup_Node_Data
     (Id   : Node_Type_Id;
      Name : String) return Any_Node_Data_Reference
   is
      Cursor : Any_Node_Type_Id := Id;
   begin
      --  Go through the derivation chain for Id and look for any field or
      --  property whose name matches Name.

      while Cursor /= None loop
         declare
            Node_Desc : Node_Type_Descriptor renames
               Node_Type_Descriptors (Cursor).all;
         begin
            for F of Node_Desc.Fields loop
               pragma Warnings (Off, "value not in range of type");
               if Field_Name (F.Field) = Name then
                  return F.Field;
               end if;
               pragma Warnings (On, "value not in range of type");
            end loop;

            for P of Node_Desc.Properties loop
               if Property_Name (P) = Name then
                  return P;
               end if;
            end loop;

            Cursor := Node_Desc.Base_Type;
         end;
      end loop;
      return None;
   end Lookup_Node_Data;

   ----------------
   -- Field_Name --
   ----------------

   function Field_Name (Field : Field_Reference) return String is
   begin
      pragma Warnings (Off, "value not in range of subtype");
      return Syntax_Field_Descriptors (Field).Name;
      pragma Warnings (On, "value not in range of subtype");
   end Field_Name;

   ----------------
   -- Field_Type --
   ----------------

   function Field_Type (Field : Field_Reference) return Node_Type_Id is
   begin
      pragma Warnings (Off, "value not in range of subtype");
      return Syntax_Field_Descriptors (Field).Field_Type;
      pragma Warnings (On, "value not in range of subtype");
   end Field_Type;

   ----------------
   -- Eval_Field --
   ----------------

   function Eval_Field
     (Node  : ${T.entity.api_name}'Class;
      Field : Field_Reference) return ${T.entity.api_name}
   is
      Kind : constant ${root_node_kind_name} := Node.Kind;
   begin
      <%
         def get_actions(astnode, node_expr):
            fields = astnode.get_parse_fields(
               predicate=lambda f: not f.overriding,
               include_inherited=False)
            result = []

            if fields:
               result.append('case Field is')
               for f in fields:
                  result.append('when {} => return {}.{}'.format(
                     f.introspection_enum_literal,
                     node_expr,
                     f.name))
                  if not f.type.is_root_node:
                     result[-1] += '.As_{}'.format(T.entity.api_name)
                  result[-1] += ';'
               result.append('when others => null;')
               result.append('end case;')

            return '\n'.join(result)
      %>
      ${ctx.generate_actions_for_hierarchy('Node', 'Kind', get_actions,
                                           public_nodes=True)}

      ## If we haven't matched the requested field on Node, report an error
      return (raise Node_Data_Evaluation_Error
              with "no such field on this node");
   end Eval_Field;

   -----------
   -- Index --
   -----------

   function Index
     (Kind : ${root_node_kind_name}; Field : Field_Reference) return Positive
   is
   begin
      % if ctx.sorted_parse_fields:
         <%
            concrete_astnodes = [n for n in ctx.astnode_types
                                 if not n.abstract]
            def enum_literal(f):
               return (f.overriding or f).introspection_enum_literal
         %>
         case Kind is
            % for n in concrete_astnodes:
               when ${n.ada_kind_name} =>
               return (case Field is
                       % for f in n.get_parse_fields( \
                          predicate=lambda f: not f.null \
                       ):
                       when ${enum_literal(f)} => ${f.index + 1},
                       % endfor
                       when others => raise Constraint_Error);
            % endfor
         end case;

      % else:
         return (raise Program_Error);
      % endif
   end Index;

   --------------------------------
   -- Field_Reference_From_Index --
   --------------------------------

   function Field_Reference_From_Index
     (Kind : ${root_node_kind_name}; Index : Positive) return Field_Reference
   is
   begin
      <%
         def get_actions(astnode, node_expr):
            fields = astnode.get_parse_fields(
               predicate=lambda f: not f.abstract and not f.null,
               include_inherited=False
            )
            result = []

            # List types have no field, so just raise an error if a list kind
            # is passed.
            if astnode.is_generic_list_type:
               result.append(
                  'raise Invalid_Field with "List AST nodes have no field";'
               )
            elif astnode.is_list:
               pass

            # Otherwise, dispatch on the index to return the corresponding
            # field enum value.
            elif fields:
               result.append('case Index is')
               for f in fields:
                  result.append('when {} => return {};'.format(
                     f.index + 1,
                     (f.overriding or f).introspection_enum_literal
                  ))
               result.append('when others => null;')
               result.append('end case;')

            return '\n'.join(result)
      %>
      ${ctx.generate_actions_for_hierarchy(None, 'Kind', get_actions)}

      pragma Warnings (Off, "value not in range of type");
      return (raise Invalid_Field with "Index is out of bounds");
      pragma Warnings (On, "value not in range of type");
   end Field_Reference_From_Index;

   ------------
   -- Fields --
   ------------

   function Fields (Kind : ${root_node_kind_name}) return Field_Reference_Array
   is
   begin
      % if ctx.sorted_parse_fields:
         return Fields (Id_For_Kind (Kind), Concrete_Only => True);
      % else:
         ${return_program_error()}
      % endif
   end Fields;

   ------------
   -- Fields --
   ------------

   function Fields
     (Id : Node_Type_Id; Concrete_Only : Boolean) return Field_Reference_Array
   is
      Cursor : Any_Node_Type_Id := Id;

      Added_Fields : array (Field_Reference) of Boolean := (others => False);
      --  Set of field references that were added to Result

      Result : Field_Reference_Array (1 .. Added_Fields'Length);
      --  Temporary to hold the result. We return Result (1 .. Last).

      Last : Natural := 0;
      --  Index of the last element in Result to return
   begin
      % if ctx.sorted_parse_fields:

         --  Go through the derivation chain for Id and collect fields. Do
         --  it in reverse order as we process base types last.
         while Cursor /= None loop
            declare
               Node_Desc : Node_Type_Descriptor renames
                  Node_Type_Descriptors (Cursor).all;
            begin
               for Field_Index in reverse Node_Desc.Fields'Range loop
                  declare
                     Field_Desc : Node_Field_Descriptor renames
                        Node_Desc.Fields (Field_Index).all;
                     Field      : Field_Reference renames Field_Desc.Field;
                  begin
                     --  Abstract fields share the same Field_Reference value
                     --  with the corresponding concrete fields, so collect
                     --  fields only once. We process fields in reverse order,
                     --  so we know that concrete ones will be processed before
                     --  the abstract fields they override.
                     if not (Concrete_Only
                             and then Field_Desc.Is_Abstract_Or_Null)
                        and then not Added_Fields (Field)
                     then
                        Added_Fields (Field) := True;
                        Last := Last + 1;
                        Result (Last) := Field;
                     end if;
                  end;
               end loop;
               Cursor := Node_Desc.Base_Type;
            end;
         end loop;

         --  At this point, Result contains elements in the opposite order as
         --  expected, so reverse it.

         for I in 1 .. Last / 2 loop
            declare
               Other_I : constant Positive := Last - I + 1;
               Swap    : constant Field_Reference := Result (I);
            begin
               Result (I) := Result (Other_I);
               Result (Other_I) := Swap;
            end;
         end loop;

         return Result (1 .. Last);

      % else:
         ${return_program_error()}
      % endif
   end Fields;

   ------------
   -- Fields --
   ------------

   function Fields (Id : Node_Type_Id) return Field_Reference_Array is
   begin
      return Fields (Id, Concrete_Only => False);
   end Fields;

   % if ctx.sorted_properties:

   -------------------
   -- Property_Name --
   -------------------

   function Property_Name (Property : Property_Reference) return String is
   begin
      return Property_Descriptors (Property).Name;
   end Property_Name;

   --------------------------
   -- Property_Return_Type --
   --------------------------

   function Property_Return_Type
     (Property : Property_Reference) return Value_Constraint is
   begin
      return Property_Descriptors (Property).Return_Type;
   end Property_Return_Type;

   -----------------------------
   -- Property_Argument_Types --
   -----------------------------

   function Property_Argument_Types
     (Property : Property_Reference) return Value_Constraint_Array is
   begin
      return Property_Descriptors (Property).Argument_Types;
   end Property_Argument_Types;

   ----------------------------
   -- Property_Argument_Name --
   ----------------------------

   function Property_Argument_Name
     (Property : Property_Reference; Argument_Number : Positive) return String
   is
   begin
      return Property_Descriptors (Property)
             .Argument_Names (Argument_Number).all;
   end Property_Argument_Name;

   -------------------
   -- Eval_Property --
   -------------------

   function Eval_Property
     (Node      : ${T.entity.api_name}'Class;
      Property  : Property_Reference;
      Arguments : Value_Array) return Value_Type
   is
      Kind   : constant ${root_node_kind_name} := Node.Kind;
      Desc   : Property_Descriptor renames Property_Descriptors (Property).all;
      Result : Any_Value_Type := No_Value;
   begin
      --  First, check that arguments match the property signature

      if Arguments'Length /= Desc.Arity then
         raise Node_Data_Evaluation_Error with "invalid number of arguments";
      end if;

      for I in Desc.Argument_Types'Range loop
         declare
            Arg : Value_Type renames Arguments (I - 1 + Arguments'First);
         begin
            if not Satisfies (Arg, Desc.Argument_Types (I)) then
               raise Node_Data_Evaluation_Error with
                  "invalid type for argument " & Desc.Argument_Names (I).all;
            end if;
         end;
      end loop;

      --  Now, we can proceed with the property evaluation

      <%
         def add_property_actions(result, node_expr, p):
            """
            Append to `result` the Ada code to call the `p` property on the
            `node_expr` node.
            """
            if p.arguments:
               # Unwrap arguments in local variables (convert Value_Type to
               # contrete types).
               result.append('declare')
               for i, arg in enumerate(p.arguments):
                  t = arg.public_type
                  result.append('{} : constant {} :='
                                .format(arg.name, t.api_name))
                  result.append("As_{} (Arguments (Arguments'First + {}))"
                                .format(t.introspection_prefix, i))

                  # Make sure we convert nodes to the precise type
                  if t.is_entity_type and t != T.entity:
                     result[-1] += '.As_{}'.format(t.api_name)

                  result[-1] += ';'

               result.append('begin')

            # Format a call to the property itself, passing arguments if there
            # are some.
            property_call = '{}.{}{}'.format(
               node_expr, p.name,
               (' ({})'.format(', '.join(str(arg.name) for arg in p.arguments))
                if p.arguments else ''))

            # Wrap this call to convert its result (concrete type to
            # Value_Type).
            result.append('Result := Create_{} ({});'
                          .format(p.type.introspection_prefix, property_call))

            if p.arguments:
               result.append('end;')

         def get_actions(astnode, node_expr):
            properties = [
                (p.base if p.overriding else p)
                for p in astnode.get_properties(
                   predicate=lambda p: p.is_public,
                   include_inherited=False)]
            result = []

            if properties:
               result.append('case Property is')
               for p in properties:
                  result.append('when {} =>'
                                .format(p.introspection_enum_literal))
                  add_property_actions(result, node_expr, p)
               result.append('when others => null;')
               result.append('end case;')

            return '\n'.join(result)
      %>
      ${ctx.generate_actions_for_hierarchy('Node', 'Kind', get_actions,
                                           public_nodes=True)}

      ## If we haven't matched the requested field on Node, report an error,
      ## otherwise, return the property result.
      if Result = No_Value then
         raise Node_Data_Evaluation_Error with "no such field on this node";
      end if;
      return Result;
   end Eval_Property;

   ----------------
   -- Properties --
   ----------------

   function Properties
     (Kind : ${root_node_kind_name}) return Property_Reference_Array is
   begin
      return Properties (Id_For_Kind (Kind));
   end Properties;

   ----------------
   -- Properties --
   ----------------

   function Properties (Id : Node_Type_Id) return Property_Reference_Array is
      Cursor : Any_Node_Type_Id := Id;

      Result : Property_Reference_Array (1 .. Property_Descriptors'Length);
      --  Temporary to hold the result. We return Result (1 .. Last).

      Last : Natural := 0;
      --  Index of the last element in Result to return
   begin
      --  Go through the derivation chain for Id and collect properties. Do
      --  it in reverse order as we process base types last.

      while Cursor /= None loop
         declare
            Node_Desc : Node_Type_Descriptor renames
               Node_Type_Descriptors (Cursor).all;
         begin
            for Prop_Desc of reverse Node_Desc.Properties loop
               Last := Last + 1;
               Result (Last) := Prop_Desc;
            end loop;
            Cursor := Node_Desc.Base_Type;
         end;
      end loop;

      --  At this point, Result contains elements in the opposite order as
      --  expected, so reverse it.

      for I in 1 .. Last / 2 loop
         declare
            Other_I : constant Positive := Last - I + 1;
            Swap    : constant Property_Reference := Result (I);
         begin
            Result (I) := Result (Other_I);
            Result (Other_I) := Swap;
         end;
      end loop;

      return Result (1 .. Last);
   end Properties;

   % endif

   ---------------------
   -- Token_Node_Kind --
   ---------------------

   function Token_Node_Kind
     (Kind : ${root_node_kind_name}) return Token_Kind
   is
      <% token_nodes = [n for n in ctx.astnode_types
                        if not n.abstract and n.is_token_node] %>
   begin
      % if ctx.generate_unparser:
         case Kind is
            % for n in token_nodes:
               when ${n.ada_kind_name} =>
                  return ${n.token_kind.ada_name};
            % endfor

            when others =>
               --  Kind is not a token node, and thus the precondition does not
               --  hold.
               return (raise Program_Error);
         end case;

      % else:
         pragma Unreferenced (Kind);
         return (raise Program_Error);
      % endif
   end Token_Node_Kind;

   ------------
   -- Adjust --
   ------------

   overriding procedure Adjust (Self : in out Value_Access_Wrapper) is
   begin
      if Self.Value = null then
         return;
      end if;

      declare
         Rec : Value_Record renames Self.Value.all;
      begin
         Rec.Ref_Count := Rec.Ref_Count + 1;
      end;
   end Adjust;

   --------------
   -- Finalize --
   --------------

   overriding procedure Finalize (Self : in out Value_Access_Wrapper) is
   begin
      if Self.Value = null then
         return;
      end if;

      --  If Self is non-null, decrement the reference count of the referenced
      --  value.

      declare
         Rec : Value_Record renames Self.Value.all;
      begin
         Rec.Ref_Count := Rec.Ref_Count - 1;

         if Rec.Ref_Count = 0 then
            --  Reference count dropped to 0: time to free the value and what
            --  is inside.

            case Rec.Kind is
               % for t in ctx.array_types:
                  % if t.exposed:
                     when ${t.introspection_kind} =>
                        Free (Rec.${t.introspection_kind});
                  % endif
               % endfor
               when others => null;
            end case;

            Free (Self.Value);
         end if;
      end;
   end Finalize;

begin
   for D in Node_Type_Descriptors'Range loop
      DSL_Name_To_Node_Type.Insert
        (Node_Type_Descriptors (D).DSL_Name, D);
   end loop;
end ${ada_lib_name}.Introspection;
